# distutils: language = c++
# cython: language_level = 3

from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.string cimport string
from cython.operator cimport dereference as deref, preincrement as preinc
from cpython cimport array

import numpy as np


cdef extern from '<iostream>' namespace 'std':
    cdef cppclass istream:
        pass
    cdef cppclass ostream:
        pass

cdef extern from '<sstream>' namespace 'std':
    cdef cppclass istringstream(istream):
        istringstream(const string& str) except +
    cdef cppclass ostringstream(ostream):
        ostringstream() except +
        string str()

cdef extern from 'LRSpline/HashSet.h':
    cdef cppclass HashSet_iterator[T]:
        T operator*()
        HashSet_iterator[T] operator++()
        bool equal(HashSet_iterator[T])

cdef extern from 'LRSpline/Basisfunction.h' namespace 'LR':
    cdef cppclass Basisfunction_ 'LR::Basisfunction':
        int getId()
        void getControlPoint(vector[double]&)
        int nVariate() const
        double evaluate(double u, double v, bool u_from_right, bool v_from_right) const
        void evaluate(vector[double]& results, double u, double v, int derivs, bool u_from_right, bool v_from_right) const
        double evaluate(double u, double v, double w, bool u_from_right, bool v_from_right, bool w_from_right) const
        void evaluate(vector[double]& results, double u, double v, double w, int derivs, bool u_from_right, bool v_from_right, bool w_from_right) const
        vector[double]& getknots(int i)
        vector[Element_*].iterator supportedElementBegin()
        vector[Element_*].iterator supportedElementEnd()

cdef extern from 'LRSpline/Element.h' namespace 'LR':
    cdef cppclass Element_ 'LR::Element':
        int getId()
        int getDim()
        double getParmin(int)
        double getParmax(int)
        int nBasisFunctions()
        HashSet_iterator[Basisfunction_*] supportBegin()
        HashSet_iterator[Basisfunction_*] supportEnd()

cdef extern from 'LRSpline/Meshline.h' namespace 'LR':
    cdef cppclass Meshline_ 'LR::Meshline':
        bool is_spanning_u()
        double const_par_
        double start_
        double stop_
        int multiplicity_

cdef extern from 'LRSpline/MeshRectangle.h' namespace 'LR':
    cdef cppclass MeshRectangle_ 'LR::MeshRectangle':
        MeshRectangle_(double u0, double v0, double w0, double u1, double v1, double w1, int multiplicity) except +
        int constDirection()
        double constParameter()
        vector[double] start_
        vector[double] stop_
        int multiplicity_
        MeshRectangle_* copy()

cdef extern from 'LRSpline/LRSpline.h':
    cdef enum refinementStrategy_ 'refinementStrategy':
        LR_MINSPAN
        LR_FULLSPAN
        LR_STRUCTURED_MESH

cdef extern from 'LRSpline/LRSpline.h' namespace 'LR':
    cdef enum parameterEdge_ 'LR::parameterEdge':
        NONE
        WEST
        EAST
        SOUTH
        NORTH
        TOP
        BOTTOM
    cdef cppclass LRSpline_ 'LR::LRSpline':
        int dimension()
        int nVariate()
        int nBasisFunctions()
        int nElements()
        double startparam(int)
        double endparam(int)
        int order(int)
        void generateIDs()
        void getEdgeFunctions(vector[Basisfunction_*]& edgeFunctions, parameterEdge_ edge, int depth) const
        void getEdgeElements(vector[Element_*]& edgeElements, parameterEdge_ edge) const
        vector[Element_*].iterator elementBegin()
        vector[Element_*].iterator elementEnd()
        Element_* getElement(int i)
        Basisfunction_* getBasisfunction(int i)
        HashSet_iterator[Basisfunction_*] basisBegin()
        HashSet_iterator[Basisfunction_*] basisEnd()
        bool setControlPoints(vector[double]& cps)
        void rebuildDimension(int dimvalue)
        void refineBasisFunction(int i)
        void refineBasisFunction(const vector[int]& i)
        void refineElement(int i)
        void refineElement(const vector[int]& i)
        void refineByDimensionIncrease(const vector[double]& error, double beta)
        void setRefStrat(refinementStrategy_ strat)
        void setRefSymmetry(int symmetry)
        void setRefMultiplicity(int mult)
        void setMaxTjoints(int n)
        void setCloseGaps(bool doClose)
        void setMaxAspectRatio(double r, bool aposterioriFix)

cdef extern from 'LRSpline/LRSplineSurface.h' namespace 'LR':
    cdef cppclass LRSplineSurface_ 'LR::LRSplineSurface' (LRSpline_):
        LRSplineSurface_() except +
        LRSplineSurface_(int n1, int n2, int order_u, int order_v) except +
        LRSplineSurface_(int n1, int n2, int order_u, int order_v, double* knot1, double* knot2) except +
        LRSplineSurface_(int n1, int n2, int order_u, int order_v, double* knot1, double* knot2, double* coef, int dim) except +
        LRSplineSurface_* copy()
        void read(istream) except +
        void write(ostream) except +
        void point(vector[double]& pt, double u, double v, int iEl) const
        void point(vector[double]& pt, double u, double v, int iEl, bool u_from_right, bool v_from_right) const
        void point(vector[vector[double]]& pts, double u, double v, int derivs, int iEl) const
        void point(vector[vector[double]]& pts, double u, double v, int derivs, bool u_from_right, bool v_from_right, int iEl) const
        void getGlobalKnotVector(vector[double]& knot_u, vector[double]& knot_v) const
        void getGlobalUniqueKnotVector(vector[double]& knot_u, vector[double]& knot_v) const
        Meshline_* insert_const_u_edge(double u, double start_v, double stop_v, int multiplicity)
        Meshline_* insert_const_v_edge(double v, double start_u, double stop_u, int multiplicity)
        void writePostscriptElements(ostream, int, int, bool, vector[int]*) const
        void writePostscriptMesh(ostream, bool, vector[int]*) const
        void writePostscriptMeshWithControlPoints(ostream, int, int) const
        vector[Meshline_*].iterator meshlineBegin()
        vector[Meshline_*].iterator meshlineEnd()
        int nMeshlines() const
        Meshline_* getMeshline(int i)
        void getBezierExtraction(int iEl, vector[double]& extractionMatrix)
        int getElementContaining(double u, double v)
        double makeIntegerKnots()

cdef extern from 'LRSpline/LRSplineVolume.h' namespace 'LR':
    cdef cppclass LRSplineVolume_ 'LR::LRSplineVolume' (LRSpline_):
        LRSplineVolume_() except +
        LRSplineVolume_(int n1, int n2, int n3, int order_u, int order_v, int order_w) except +
        LRSplineVolume_(int n1, int n2, int n3, int order_u, int order_v, int order_w, double* knot1, double* knot2, double *knot3) except +
        LRSplineVolume_(int n1, int n2, int n3, int order_u, int order_v, int order_w, double* knot1, double* knot2, double *knot3, double* coef, int dim) except +
        LRSplineVolume_* copy()
        void read(istream) except +
        void write(ostream) except +
        void point(vector[double]& pt, double u, double v, double w, int iEl) const
        void point(vector[double]& pt, double u, double v, double w, int iEl, bool u_from_right, bool v_from_right, bool w_from_right) const
        void point(vector[vector[double]]& pts, double u, double v, double w, int derivs, int iEl) const
        void point(vector[vector[double]]& pts, double u, double v, double w, int derivs, bool u_from_right, bool v_from_right, bool w_from_right, int iEl) const
        void getGlobalKnotVector(vector[double]& knot_u, vector[double]& knot_v, vector[double]& knot_w) const
        void getGlobalUniqueKnotVector(vector[double]& knot_u, vector[double]& knot_v, vector[double]& knot_w) const
        vector[MeshRectangle_*].iterator meshrectBegin()
        vector[MeshRectangle_*].iterator meshrectEnd()
        int nMeshRectangles() const
        MeshRectangle_* getMeshRectangle(int i)
        MeshRectangle_* insert_line(MeshRectangle_* newRect)
        void getBezierExtraction(int iEl, vector[double]& extractionMatrix)
        int getElementContaining(double u, double v, double w)


def _is_start(stream):
    peek = stream.peek(20)
    if not peek:
        return True
    if isinstance(peek, str):
        return peek.startswith('# LRSPLINE')
    return peek.startswith(b'# LRSPLINE')


class EOFError(Exception):
    pass


def _read_object(stream):
    lines = [stream.readline()]
    while not _is_start(stream):
        lines.append(stream.readline())
    if len(lines) <= 1:
        raise EOFError('No LRSpline object found')
    if isinstance(lines[0], str):
        return ''.join(lines)
    return b''.join(lines)


cdef class Basisfunction:

    cdef Basisfunction_* w

    def getId(self):
        return self.w.getId()

    def getControlPoint(self):
        cdef vector[double] data
        self.w.getControlPoint(data)
        return np.array(data)

    def nVariate(self):
        return self.w.nVariate()

    def evaluate(self, *args):
        cdef vector[double] results
        if len(args) == 4:
            u, v, ufr, vfr = args
            return self.w.evaluate(<double> u, <double> v, <bool> ufr, <bool> vfr)
        if len(args) == 5:
            u, v, d, ufr, vfr = args
            self.w.evaluate(results, <double> u, <double> v, <int> d, <bool> ufr, <bool> vfr)
            return np.array(results)
        if len(args) == 6:
            u, v, w, ufr, vfr, wfr = args
            return self.w.evaluate(<double> u, <double> v, <double> w, <bool> ufr, <bool> vfr, <bool> wfr)
        if len(args) == 7:
            u, v, w, d, ufr, vfr, wfr = args
            self.w.evaluate(results, <double> u, <double> v, <double> w, <int> d, <bool> ufr, <bool> vfr, <bool> wfr)
            return np.array(results)
        raise TypeError("evaluate() expected 4, 5, 6 or 7 arguments")

    def getknots(self, idx):
        return np.array(self.w.getknots(idx))

    def supportIter(self):
        cdef vector[Element_*].iterator it = self.w.supportedElementBegin()
        cdef vector[Element_*].iterator end = self.w.supportedElementEnd()
        while it != end:
            el = Element()
            el.w = deref(it)
            yield el
            preinc(it)


cdef class Element:

    cdef Element_* w

    def getId(self):
        return self.w.getId()

    def getDim(self):
        return self.w.getDim()

    def getParmin(self, i):
        return self.w.getParmin(i)

    def getParmax(self, i):
        return self.w.getParmax(i)

    def nBasisFunctions(self):
        return self.w.nBasisFunctions()

    def supportIter(self):
        cdef HashSet_iterator[Basisfunction_*] it = self.w.supportBegin()
        cdef HashSet_iterator[Basisfunction_*] end = self.w.supportEnd()
        while not it.equal(end):
            bf = Basisfunction()
            bf.w = deref(it)
            yield bf
            preinc(it)


cdef class Meshline:

    cdef Meshline_* w

    def is_spanning_u(self):
        return self.w.is_spanning_u()

    @property
    def const_par_(self):
        return self.w.const_par_

    @property
    def start_(self):
        return self.w.start_

    @property
    def stop_(self):
        return self.w.stop_

    @property
    def multiplicity_(self):
        return self.w.multiplicity_


cdef class MeshRectangle:

    cdef MeshRectangle_* w
    cdef bool owned

    def __cinit__(self, *args):
        if len(args) >= 6:
            multiplicity = 1 if len(args) < 7 else args[6]
            self.w = new MeshRectangle_(args[0], args[1], args[2], args[3], args[4], args[5], multiplicity)
            self.owned = True
        else:
            assert len(args) == 0
            self.w = NULL
            self.owned = False

    cdef set_wrappee(self, MeshRectangle_* w, bool owned):
        if self.owned:
            del self.w
        self.w = w
        self.owned = owned

    def __dealloc__(self):
        if self.owned:
            del self.w

    def constDirection(self):
        return self.w.constDirection()

    def constParameter(self):
        return self.w.constParameter()

    @property
    def start_(self):
        return np.array(self.w.start_)

    @property
    def stop_(self):
        return np.array(self.w.stop_)

    @property
    def multiplicity_(self):
        return self.w.multiplicity_

    def copy(self):
        return MeshRectangle(*self.start_, *self.stop_, self.multiplicity_)


cdef class parameterEdge:
    NONE   = parameterEdge_.NONE
    WEST   = parameterEdge_.WEST
    EAST   = parameterEdge_.EAST
    SOUTH  = parameterEdge_.SOUTH
    NORTH  = parameterEdge_.NORTH
    TOP    = parameterEdge_.TOP
    BOTTOM = parameterEdge_.BOTTOM


cdef class refinementStrategy:
    LR_MINSPAN = refinementStrategy_.LR_MINSPAN
    LR_FULLSPAN = refinementStrategy_.LR_FULLSPAN
    LR_STRUCTURED_MESH = refinementStrategy_.LR_STRUCTURED_MESH


cdef class LRSplineObject:

    cdef LRSpline_* w

    def __dealloc__(self):
        del self.w

    def nVariate(self):
        return self.w.nVariate()

    def dimension(self):
        return self.w.dimension()

    def rebuildDimension(self, int dim):
        self.w.rebuildDimension(dim)

    def nBasisFunctions(self):
        return self.w.nBasisFunctions()

    def nElements(self):
        return self.w.nElements()

    def startparam(self, int i):
        return self.w.startparam(i)

    def endparam(self, int i):
        return self.w.endparam(i)

    def order(self, int i):
        return self.w.order(i)

    def elementIter(self):
        cdef vector[Element_*].iterator it = self.w.elementBegin()
        cdef vector[Element_*].iterator end = self.w.elementEnd()
        while it != end:
            el = Element()
            el.w = deref(it)
            yield el
            preinc(it)

    def getElement(self, int i):
        el = Element()
        el.w = self.w.getElement(i)
        return el

    def getEdgeElementsIter(self, edge):
        cdef vector[Element_*] els
        self.w.getEdgeElements(els, edge)
        it = els.begin()
        while it != els.end():
            bf = Element()
            bf.w = deref(it)
            yield bf
            preinc(it)

    def basisIter(self):
        cdef HashSet_iterator[Basisfunction_*] it = self.w.basisBegin()
        cdef HashSet_iterator[Basisfunction_*] end = self.w.basisEnd()
        while not it.equal(end):
            bf = Basisfunction()
            bf.w = deref(it)
            yield bf
            preinc(it)

    def getBasisfunction(self, int i):
        bf = Basisfunction()
        bf.w = self.w.getBasisfunction(i)
        return bf

    def getEdgeFunctionsIter(self, edge, int depth=1):
        cdef vector[Basisfunction_*] bfs
        self.w.getEdgeFunctions(bfs, edge, depth)
        it = bfs.begin()
        while it != bfs.end():
            bf = Basisfunction()
            bf.w = deref(it)
            yield bf
            preinc(it)

    def setControlpoints(self, cpts):
        self.w.setControlPoints(cpts)

    def generateIDs(self):
        self.w.generateIDs()

    def refineBasisFunction(self, i):
        cdef vector[int] indices
        if isinstance(i, int):
            self.w.refineBasisFunction(<int> i)
        elif isinstance(i, list):
            indices = i
            self.w.refineBasisFunction(indices)

    def refineElement(self, i):
        cdef vector[int] indices
        if isinstance(i, int):
            self.w.refineElement(<int> i)
        elif isinstance(i, list):
            indices = i
            self.w.refineElement(indices)

    def refineByDimensionIncrease(self, errors, double beta):
        cdef vector[double] cpperrors
        cpperrors = errors
        self.w.refineByDimensionIncrease(errors, beta)

    def setRefStrat(self, strat):
        self.w.setRefStrat(strat)

    def setRefSymmetry(self, int symmetry):
        self.w.setRefSymmetry(symmetry)

    def setRefMultiplicity(self, int mult):
        self.w.setRefMultiplicity(mult)

    def setMaxTjoints(self, int n):
        self.w.setMaxTjoints(n)

    def setCloseGaps(self, bool close):
        self.w.setCloseGaps(close)

    def setMaxAspectRatio(self, double r, bool aposteriori=True):
        self.w.setMaxAspectRatio(r, aposteriori)


cdef class LRSurface(LRSplineObject):

    def __cinit__(self, n1=None, n2=None, order_u=None, order_v=None, knot1=None, knot2=None, coef=None, dim=2):
        cdef array.array knot1_arr
        cdef array.array knot2_arr
        cdef array.array coef_arr
        if knot1 is not None and knot2 is not None and coef is not None:
            knot1_arr = array.array('d', knot1)
            knot2_arr = array.array('d', knot2)
            coef_arr  = array.array('d', coef)
            self.w = new LRSplineSurface_(n1, n2, order_u, order_v, knot1_arr.data.as_doubles, knot2_arr.data.as_doubles, coef_arr.data.as_doubles, dim)
        elif knot1 is not None and knot2 is not None:
            knot1_arr = array.array('d', knot1)
            knot2_arr = array.array('d', knot2)
            self.w = new LRSplineSurface_(n1, n2, order_u, order_v, knot1_arr.data.as_doubles, knot2_arr.data.as_doubles)
            if dim != 2:
                self.w.rebuildDimension(dim)
        elif n1!=None and n2!=None and order_u!=None and order_v!=None:
            self.w = new LRSplineSurface_(n1, n2, order_u, order_v)
            if dim != 2:
                self.w.rebuildDimension(dim)
        else:
            self.w = new LRSplineSurface_()

    cdef _set_w(self, LRSplineSurface_* w):
        del self.w
        self.w = w

    def read(self, stream):
        cdef string cppstring
        if hasattr(stream, 'readline'):
            stream = _read_object(stream)
        if isinstance(stream, str):
            cppstring = stream.encode()
        elif isinstance(stream, bytes):
            cppstring = stream
        cdef istringstream* cppstream
        cppstream = new istringstream(cppstring)
        (<LRSplineSurface_*> self.w).read(deref(cppstream))
        del stream

    def write(self, stream):
        cdef ostringstream* cppstream
        cppstream = new ostringstream()
        (<LRSplineSurface_*> self.w).write(deref(cppstream))
        stream.write(cppstream.str())
        del cppstream

    def writePostscriptElements(self, stream, int nu=2, int nv=2, bool close=True, colorElements=None):
        cdef ostringstream* cppstream
        cdef vector[int] colors
        cppstream = new ostringstream()
        if colorElements is None:
            (<LRSplineSurface_*> self.w).writePostscriptElements(deref(cppstream), nu, nv, close, NULL)
        else:
            colors = colorElements
            (<LRSplineSurface_*> self.w).writePostscriptElements(deref(cppstream), nu, nv, close, &colors)
        stream.write(cppstream.str())
        del cppstream

    def copy(self):
        cdef LRSplineSurface_* copy = (<LRSplineSurface_*> self.w).copy()
        retval = LRSurface()
        retval._set_w(copy)
        return retval

    def getGlobalKnotVector(self):
        cdef vector[double] knots_u
        cdef vector[double] knots_v
        (<LRSplineSurface_*> self.w).getGlobalKnotVector(knots_u, knots_v)
        return np.array(knots_u), np.array(knots_v)

    def getGlobalUniqueKnotVector(self):
        cdef vector[double] knots_u
        cdef vector[double] knots_v
        (<LRSplineSurface_*> self.w).getGlobalUniqueKnotVector(knots_u, knots_v)
        return np.array(knots_u), np.array(knots_v)

    def point(self, *args, iEl=-1):
        cdef vector[double] point
        cdef vector[vector[double]] results
        if len(args) == 2:
            u, v = args
            (<LRSplineSurface_*> self.w).point(point, u, v, <int> iEl)
            return np.array(point)
        if len(args) == 4:
            u, v, ufr, vfr = args
            (<LRSplineSurface_*> self.w).point(point, u, v, <int> iEl, <bool> ufr, <bool> vfr)
            return np.array(point)
        if len(args) == 3:
            u, v, d = args
            (<LRSplineSurface_*> self.w).point(results, u, v, d, <int> iEl)
            return np.array(results)
        if len(args) == 5:
            u, v, d, ufr, vfr = args
            (<LRSplineSurface_*> self.w).point(results, u, v, d, <bool> ufr, <bool> vfr, <int> iEl)
            return np.array(results)
        raise TypeError("point() expected 2, 3, 4 or 5 arguments")

    def meshlineIter(self):
        cdef vector[Meshline_*].iterator it = (<LRSplineSurface_*> self.w).meshlineBegin()
        cdef vector[Meshline_*].iterator end = (<LRSplineSurface_*> self.w).meshlineEnd()
        while it != end:
            ml = Meshline()
            ml.w = deref(it)
            yield ml
            preinc(it)

    def nMeshlines(self):
        return (<LRSplineSurface_*> self.w).nMeshlines()

    def getMeshline(self, int i):
        ml = Meshline()
        ml.w = (<LRSplineSurface_*> self.w).getMeshline(i)
        return ml

    def insert_const_u_edge(self, double u, double start_v, double stop_v, int multiplicity=1):
        (<LRSplineSurface_*> self.w).insert_const_u_edge(u, start_v, stop_v, multiplicity)

    def insert_const_v_edge(self, double v, double start_u, double stop_u, int multiplicity=1):
        (<LRSplineSurface_*> self.w).insert_const_v_edge(v, start_u, stop_u, multiplicity)

    def getGlobalUniqueKnotVector(self):
        cdef vector[double] ktsu
        cdef vector[double] ktsv
        (<LRSplineSurface_*> self.w).getGlobalUniqueKnotVector(ktsu, ktsv)
        return (np.array(ktsu), np.array(ktsv))

    def getBezierExtraction(self, iEl=-1):
        cdef vector[double] result
        width  = self.w.order(0) * self.w.order(1)
        height = self.w.getElement(iEl).nBasisFunctions()
        (<LRSplineSurface_*> self.w).getBezierExtraction(iEl, result)
        return np.reshape(result, (height, width), order='F')

    def getElementContaining(self, double u, double v):
        return (<LRSplineSurface_*> self.w).getElementContaining(u,v)

    def makeIntegerKnots(self):
        return (<LRSplineSurface_*> self.w).makeIntegerKnots()


cdef class LRVolume(LRSplineObject):

    def __cinit__(self, n1=None, n2=None, n3=None, order_u=None, order_v=None, order_w=None, knot1=None, knot2=None, knot3=None, coef=None, dim=3):
        cdef array.array knot1_arr
        cdef array.array knot2_arr
        cdef array.array knot3_arr
        cdef array.array coef_arr
        if knot1 is not None and knot2 is not None and knot3 is not None and coef is not None:
            knot1_arr = array.array('d', knot1)
            knot2_arr = array.array('d', knot2)
            knot3_arr = array.array('d', knot3)
            coef_arr  = array.array('d', coef)
            self.w = new LRSplineVolume_(n1, n2, n3, order_u, order_v, order_w, knot1_arr.data.as_doubles, knot2_arr.data.as_doubles, knot3_arr.data.as_doubles, coef_arr.data.as_doubles, dim)
        elif knot1 is not None and knot2 is not None and knot3 is not None:
            knot1_arr = array.array('d', knot1)
            knot2_arr = array.array('d', knot2)
            knot3_arr = array.array('d', knot3)
            self.w = new LRSplineVolume_(n1, n2, n3, order_u, order_v, order_w, knot1_arr.data.as_doubles, knot2_arr.data.as_doubles, knot3_arr.data.as_doubles)
            if dim != 3:
                self.w.rebuildDimension(dim)
        elif n1!=None and n2!=None and n3!=None and order_u!=None and order_v!=None and order_w!=None:
            self.w = new LRSplineVolume_(n1, n2, n3, order_u, order_v, order_w)
            if dim != 3:
                self.w.rebuildDimension(dim)
        else:
            self.w = new LRSplineVolume_()

    cdef _set_w(self, LRSplineVolume_* w):
        del self.w
        self.w = w

    def read(self, stream):
        cdef string cppstring
        if hasattr(stream, 'readline'):
            stream = _read_object(stream)
        if isinstance(stream, str):
            cppstring = stream.encode()
        elif isinstance(stream, bytes):
            cppstring = stream
        cdef istringstream* cppstream
        cppstream = new istringstream(cppstring)
        (<LRSplineVolume_*> self.w).read(deref(cppstream))
        del stream

    def write(self, stream):
        cdef ostringstream* cppstream
        cppstream = new ostringstream()
        (<LRSplineVolume_*> self.w).write(deref(cppstream))
        stream.write(cppstream.str())
        del cppstream

    def copy(self):
        cdef LRSplineVolume_* copy = (<LRSplineVolume_*> self.w).copy()
        retval = LRVolume()
        retval._set_w(copy)
        return retval

    def getGlobalKnotVector(self):
        cdef vector[double] knots_u
        cdef vector[double] knots_v
        cdef vector[double] knots_w
        (<LRSplineVolume_*> self.w).getGlobalKnotVector(knots_u, knots_v, knots_w)
        return np.array(knots_u), np.array(knots_v), np.array(knots_w)

    def getGlobalUniqueKnotVector(self):
        cdef vector[double] knots_u
        cdef vector[double] knots_v
        cdef vector[double] knots_w
        (<LRSplineVolume_*> self.w).getGlobalUniqueKnotVector(knots_u, knots_v, knots_w)
        return np.array(knots_u), np.array(knots_v), np.array(knots_w)

    def point(self, *args, iEl=-1):
        cdef vector[double] point
        cdef vector[vector[double]] results
        if len(args) == 3:
            u, v, w = args
            (<LRSplineVolume_*> self.w).point(point, u, v, w, <int> iEl)
            return np.array(point)
        if len(args) == 6:
            u, v, w, ufr, vfr, wfr = args
            (<LRSplineVolume_*> self.w).point(point, u, v, w, <int> iEl, <bool> ufr, <bool> vfr, <bool> wfr)
            return np.array(point)
        if len(args) == 4:
            u, v, w, d = args
            (<LRSplineVolume_*> self.w).point(results, u, v, w, d, <int> iEl)
            return np.array(results)
        if len(args) == 7:
            u, v, w, d, ufr, vfr, wfr = args
            (<LRSplineVolume_*> self.w).point(results, u, v, w, d, <bool> ufr, <bool> vfr, <bool> wfr, <int> iEl)
            return np.array(results)
        raise TypeError("point() expected 2, 3, 4 or 5 arguments")

    def meshrectIter(self):
        cdef vector[MeshRectangle_*].iterator it = (<LRSplineVolume_*> self.w).meshrectBegin()
        cdef vector[MeshRectangle_*].iterator end = (<LRSplineVolume_*> self.w).meshrectEnd()
        while it != end:
            mr = MeshRectangle()
            mr.set_wrappee(deref(it), False)
            yield mr
            preinc(it)

    def nMeshRectangles(self):
        return (<LRSplineVolume_*> self.w).nMeshRectangles()

    def getMeshRectangle(self, i):
        mr = MeshRectangle()
        mr.set_wrappee((<LRSplineVolume_*> self.w).getMeshRectangle(i), False)
        return mr

    def insert_line(self, mr: MeshRectangle):
        (<LRSplineVolume_*> self.w).insert_line(mr.w.copy())

    def getGlobalUniqueKnotVector(self):
        cdef vector[double] ktsu
        cdef vector[double] ktsv
        cdef vector[double] ktsw
        (<LRSplineVolume_*> self.w).getGlobalUniqueKnotVector(ktsu, ktsv, ktsw)
        return (np.array(ktsu), np.array(ktsv), np.array(ktsw))

    def getBezierExtraction(self, iEl=-1):
        cdef vector[double] result
        width  = self.w.order(0) * self.w.order(1) * self.w.order(2)
        height = self.w.getElement(iEl).nBasisFunctions()
        (<LRSplineVolume_*> self.w).getBezierExtraction(iEl, result)
        return np.reshape(result, (height, width))

    def getElementContaining(self, double u, double v, double w):
        return (<LRSplineVolume_*> self.w).getElementContaining(u,v,w)



from __future__ import annotations

import io
import operator as op
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Iterator, Sequence
from functools import partial
from itertools import chain, combinations_with_replacement, repeat
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Generic,
    Literal,
    Protocol,
    Self,
    SupportsFloat,
    TextIO,
    TypedDict,
    TypeVar,
    Unpack,
    cast,
    overload,
)

import numpy as np
from numpy.typing import NDArray

from . import raw

if TYPE_CHECKING:
    from numpy.typing import ArrayLike

__version__ = "1.14.4"

L = TypeVar("L", bound="LRSplineObject")
R = TypeVar("R", bound=raw.LRSplineObject)
S = TypeVar("S")
T = TypeVar("T")
U = TypeVar("U")

Direction = int | Literal["u", "U", "v", "V", "w", "W"]
Edge = Literal["west", "east", "north", "south", "top", "bottom"]
FArray = NDArray[np.floating]
Scalar = float | np.floating
Scalars = Sequence[Scalar] | FArray
ScalarOrScalars = Scalar | Scalars


class ConfigureKwargs(TypedDict, total=False):
    aspect_ratio: float
    posteriori_fix: bool
    strategy: int
    symmetry: int
    multiplicity: int
    max_tjoints: int
    close_gaps: bool


class WritePsKwargs(TypedDict, total=False):
    nu: int
    nv: int
    close: bool
    colorElements: Sequence[int] | None


def _ensure_scalars(x: ScalarOrScalars | tuple[Scalar], dups: int = 1) -> list[float]:
    if isinstance(x, SupportsFloat) and not isinstance(x, np.ndarray):
        return [float(x)] * dups
    retval = list(map(float, x))
    if len(retval) < dups and retval:
        retval.extend(float(x[-1]) for _ in range(dups - len(retval)))
    return retval


def _check_direction(direction: Direction, pardim: int) -> int:
    if direction in {0, "u", "U"} and pardim > 0:
        return 0
    if direction in {1, "v", "V"} and pardim > 1:
        return 1
    if direction in {2, "w", "W"} and pardim > 2:
        return 2
    raise ValueError("Invalid direction")


def _check_edge(edge: Iterable[Edge]) -> int:
    side = raw.parameterEdge.NONE
    for arg in edge:
        side |= {
            "west": raw.parameterEdge.WEST,
            "east": raw.parameterEdge.EAST,
            "south": raw.parameterEdge.SOUTH,
            "north": raw.parameterEdge.NORTH,
            "top": raw.parameterEdge.TOP,
            "bottom": raw.parameterEdge.BOTTOM,
        }[arg]
    return side


def _constructor(stream: TextIO | BinaryIO | bytes | str) -> type[LRSplineObject]:
    if isinstance(stream, TextIO | BinaryIO):
        # if hasattr(stream, "readline"):
        peek = stream.readline()
        if not peek:
            raise raw.EOFError("")
    else:
        peek = stream
    if isinstance(peek, bytes):
        peek = peek.decode("utf-8")
    if peek.startswith("# LRSPLINE SURFACE"):
        return LRSplineSurface
    if peek.startswith("# LRSPLINE VOLUME"):
        return LRSplineVolume
    raise ValueError(f"Unknown LRSpline object type: '{peek}'")


def _derivative_index(d: tuple[int, int] | tuple[int, int, int]) -> tuple[int, int]:
    """Calculate the derivative index of 'd' (a 2-tuple or 3-tuple) using
    LRSplines' derivative numbering scheme.  Return nderivs and index.
    """
    nderivs = sum(d)
    if len(d) == 2:
        index = sum(dd + 1 for dd in range(nderivs)) + d[1]
        return nderivs, index
    index = nderivs * (nderivs + 1) * (nderivs + 2) // 6
    tgt = tuple(chain.from_iterable(repeat(i, r) for i, r in enumerate(d)))
    index += next(i for i, t in enumerate(combinations_with_replacement(range(len(d)), nderivs)) if t == tgt)
    return nderivs, index


class DerivativeHelperFunc2d(Protocol):
    def __call__(self, u: np.floating, v: np.floating, nderivs: int) -> FArray: ...


class DerivativeHelperFunc3d(Protocol):
    def __call__(self, u: np.floating, v: np.floating, w: np.floating, nderivs: int) -> FArray: ...


@overload
def _derivative_helper(
    pts: tuple[FArray, FArray] | tuple[Scalar, Scalar],
    derivs: tuple[int, int] | Sequence[tuple[int, int]],
    func: DerivativeHelperFunc2d,
) -> FArray: ...


@overload
def _derivative_helper(
    pts: tuple[FArray, FArray, FArray] | tuple[Scalar, Scalar, Scalar],
    derivs: tuple[int, int, int] | Sequence[tuple[int, int, int]],
    func: DerivativeHelperFunc3d,
) -> FArray: ...


def _derivative_helper(pts, derivs, func):  # type: ignore[no-untyped-def]
    """Helper for calculating derivatives.

    Pts must be a tuple of points, or a tuple of point arrays.

    Derivs must be a tuple indicating which derivative to take, or a
    list of such tuples.

    Func must be a callable accepting two or three floats (the point
    to evaluate) and an 'nderivs' argument, returning a 1-dimensional
    array.

    - If one point, one deriv: return a 1-dimensional array (coord)
    - If multiple points, one deriv: return a 2-dimensional array (point x coord)
    - If one point, multiple derivs: return a 2-dimensional array (deriv x coord)
    - If multiple points, multiple derivs: return 3-dimensional (deriv x point x coord)
    """

    singlept = not isinstance(pts[0], np.ndarray)

    # If requesting a single derivative
    if isinstance(derivs[0], int):
        nderiv, index = _derivative_index(derivs)

        # If requesting a single point
        if singlept:
            # pts = cast(tuple[Scalar, ...], pts)
            return cast("FArray", func(*pts, nderivs=nderiv)[index])

        # If requesting multiple points
        data_list = [func(*pt, nderivs=nderiv) for pt in zip(*pts)]
        return np.array([d[index] for d in data_list])

    # If requesting several derivatives
    nderiv, indexes = 0, []
    for deriv in derivs:
        n, i = _derivative_index(deriv)
        nderiv = max(n, nderiv)
        indexes.append(i)

    # If requesting a single point
    if singlept:
        # pts = cast(tuple[Scalar, ...], pts)
        data = func(*pts, nderivs=nderiv)
        return np.array([data[i] for i in indexes])

    # If requesting multiple points
    data_list = [func(*pt, nderivs=nderiv) for pt in zip(*pts)]
    return np.array([[d[i] for d in data_list] for i in indexes])


class SimpleWrapper(Generic[T]):
    lr: LRSplineObject
    w: T

    def __init__(self, lr: LRSplineObject, w: T) -> None:
        self.w = w
        self.lr = lr


class BasisFunction(SimpleWrapper[raw.Basisfunction]):
    def __str__(self) -> str:
        return f"BasisFunction#{self.id}"

    @property
    def id(self) -> int:
        return self.w.getId()

    @property
    def controlpoint(self) -> FArray:
        return self.w.getControlPoint()

    @property
    def nvariate(self) -> int:
        return self.w.nVariate()

    def evaluate(self, u: FArray, v: FArray) -> FArray:
        retval = np.array([self.w.evaluate(up, vp, True, True) for up, vp in zip(u.flatten(), v.flatten())])
        return retval.reshape(u.shape)

    def derivative(self, *pts: Scalar | FArray, d: tuple[int, int] | tuple[int, int, int] = (1, 1)) -> FArray:
        if self.nvariate == 2:

            def wrapper_2d(u: np.floating, v: np.floating, nderivs: int) -> FArray:
                return self.w.evaluate(u, v, nderivs, True, True)

            pts = cast("tuple[Scalar, Scalar] | tuple[FArray, FArray]", pts)
            d = cast("tuple[int, int]", d)
            return _derivative_helper(pts, d, wrapper_2d)

        def wrapper_3d(u: np.floating, v: np.floating, w: np.floating, nderivs: int) -> FArray:
            return self.w.evaluate(u, v, w, nderivs, True, True, True)

        pts = cast("tuple[Scalar, Scalar, Scalar] | tuple[FArray, FArray, FArray]", pts)
        d = cast("tuple[int, int, int]", d)
        return _derivative_helper(pts, d, wrapper_3d)

    def support(self) -> Iterator[Element]:
        for w in self.w.supportIter():
            yield Element(self.lr, w)

    def __getitem__(self, idx: int) -> FArray:
        return self.w.getknots(idx)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, BasisFunction):
            return False
        return self.id == other.id

    __call__ = evaluate

    def refine(self) -> None:
        self.lr.w.refineBasisFunction(self.id)
        self.lr.w.generateIDs()


class Element(SimpleWrapper[raw.Element]):
    def __str__(self) -> str:
        return f"Element#{self.id}"

    @property
    def id(self) -> int:
        return self.w.getId()

    @property
    def pardim(self) -> int:
        return self.w.getDim()

    @overload
    def start(self) -> tuple[float, ...]: ...

    @overload
    def start(self, direction: Direction) -> float: ...

    def start(self, direction=None):  # type: ignore[no-untyped-def]
        if direction is None:
            return tuple(self.w.getParmin(d) for d in range(self.pardim))
        return self.w.getParmin(_check_direction(direction, self.pardim))

    @overload
    def end(self) -> tuple[float, ...]: ...

    @overload
    def end(self, direction: Direction) -> float: ...

    def end(self, direction=None):  # type: ignore[no-untyped-def]
        if direction is None:
            return tuple(self.w.getParmax(d) for d in range(self.pardim))
        return self.w.getParmax(_check_direction(direction, self.pardim))

    @overload
    def span(self) -> tuple[tuple[float, float], ...]: ...

    @overload
    def span(self, direction: Direction) -> tuple[float, float]: ...

    def span(self, direction=None):  # type: ignore[no-untyped-def]
        if direction is None:
            return tuple((self.w.getParmin(d), self.w.getParmax(d)) for d in range(self.pardim))
        direction = _check_direction(direction, self.pardim)
        return (self.w.getParmin(direction), self.w.getParmax(direction))

    def support(self) -> Iterator[BasisFunction]:
        for w in self.w.supportIter():
            yield BasisFunction(self.lr, w)

    def refine(self) -> None:
        self.lr.w.refineElement(self.id)
        self.lr.w.generateIDs()

    def bezier_extraction(self) -> FArray:
        return self.lr.bezier_extraction(self.id)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Element):
            return False
        return self.id == other.id


I = TypeVar("I", raw.Meshline, raw.MeshRectangle)


class MeshInterface(ABC, SimpleWrapper[I]):
    @property
    @abstractmethod
    def constant_direction(self) -> int: ...

    @property
    @abstractmethod
    def value(self) -> float: ...

    @property
    @abstractmethod
    def variable_directions(self) -> tuple[int, ...]: ...

    @property
    @abstractmethod
    def multiplicity(self) -> int: ...

    @overload
    def span(self) -> tuple[tuple[float, float], ...]: ...

    @overload
    def span(self, direction: Direction) -> tuple[float, float]: ...

    @abstractmethod
    def span(self, direction=None):  # type: ignore[no-untyped-def]
        ...

    def __str__(self) -> str:
        cls = self.__class__.__name__
        const = "{} = {}".format("uvw"[self.constant_direction], self.value)
        variables = [
            "{} < {} < {}".format(self.start(i), "uvw"[i], self.end(i)) for i in self.variable_directions
        ]
        return cls + "(" + const + "; " + ", ".join(variables) + ")"

    @overload
    def start(self) -> tuple[float, ...]: ...

    @overload
    def start(self, direction: Direction) -> float: ...

    def start(self, direction=None):  # type: ignore[no-untyped-def]
        if direction is None:
            return tuple(self.span(d)[0] for d in range(self.lr.pardim))
        return self.span(direction)[0]

    @overload
    def end(self) -> tuple[float, ...]: ...

    @overload
    def end(self, direction: Direction) -> float: ...

    def end(self, direction=None):  # type: ignore[no-untyped-def]
        if direction is None:
            return tuple(self.span(d)[1] for d in range(self.lr.pardim))
        return self.span(direction)[1]


class MeshLine(MeshInterface[raw.Meshline]):
    @property
    def variable_direction(self) -> int:
        return 0 if self.w.is_spanning_u() else 1

    @property
    def variable_directions(self) -> tuple[int, ...]:
        return (0,) if self.w.is_spanning_u() else (1,)

    @property
    def constant_direction(self) -> int:
        return 1 if self.w.is_spanning_u() else 0

    @property
    def value(self) -> float:
        return self.w.const_par_

    @property
    def multiplicity(self) -> int:
        return self.w.multiplicity_

    def span(self, direction=None):  # type: ignore[no-untyped-def]
        if direction is None:
            var = (self.w.start_, self.w.stop_)
            const = (self.w.const_par_, self.w.const_par_)
            return (var, const) if self.w.is_spanning_u() else (const, var)
        direction = _check_direction(direction, self.lr.pardim)
        if direction == self.variable_direction:
            return (self.w.start_, self.w.stop_)
        return (self.w.const_par_, self.w.const_par_)


class MeshRect(MeshInterface[raw.MeshRectangle]):
    @property
    def variable_directions(self) -> tuple[int, ...]:
        return {
            0: (1, 2),
            1: (0, 2),
            2: (0, 1),
        }[self.w.constDirection()]

    @property
    def constant_direction(self) -> int:
        return self.w.constDirection()

    @property
    def value(self) -> float:
        return self.w.constParameter()

    @property
    def multiplicity(self) -> int:
        return self.w.multiplicity_

    def span(self, direction=None):  # type: ignore[no-untyped-def]
        start, end = self.w.start_, self.w.stop_
        if direction is None:
            return tuple(zip(start, end))
        direction = _check_direction(direction, self.lr.pardim)
        return start[direction], end[direction]


class ListLikeView(Generic[S, T, U]):
    obj: S
    lenf: Callable[[S], int]
    itemf: Callable[[S], Callable[[int], T]]
    iterf: Callable[[S], Iterator[T]]
    wrapf: Callable[[T], U]

    def __init__(self, obj: S, lenf: str, itemf: str, iterf: str, wrapf: Callable[[T], U]):
        self.obj = obj
        self.lenf = cast("Callable[[S], int]", op.methodcaller(lenf))
        self.itemf = cast("Callable[[S], Callable[[int], T]]", op.attrgetter(itemf))
        self.iterf = cast("Callable[[S], Iterator[T]]", op.methodcaller(iterf))
        self.wrapf = wrapf

    def __len__(self) -> int:
        return self.lenf(self.obj)

    def __getitem__(self, idx: int) -> U:
        return self.wrapf(self.itemf(self.obj)(idx))

    def __iter__(self) -> Iterator[U]:
        for w in self.iterf(self.obj):
            yield self.wrapf(w)


class ElementView(ListLikeView[R, raw.Element, Element]):
    lr: LRSplineObject[R]

    def __init__(self, lr: LRSplineObject[R]) -> None:
        super().__init__(lr.w, "nElements", "getElement", "elementIter", partial(Element, lr))
        self.lr = lr

    def edge(self, *edge: Edge) -> Iterator[Element]:
        for w in self.lr.w.getEdgeElementsIter(_check_edge(edge)):
            yield Element(self.lr, w)


class BasisView(ListLikeView[R, raw.Basisfunction, BasisFunction]):
    lr: LRSplineObject[R]

    def __init__(self, lr: LRSplineObject[R]) -> None:
        super().__init__(lr.w, "nBasisFunctions", "getBasisfunction", "basisIter", partial(BasisFunction, lr))
        self.lr = lr

    def edge(self, *edge: Edge, depth: int = 1) -> Iterator[BasisFunction]:
        for w in self.lr.w.getEdgeFunctionsIter(_check_edge(edge)):
            yield BasisFunction(self.lr, w)


class MeshLineView(ListLikeView[raw.LRSurface, raw.Meshline, MeshLine]):
    def __init__(self, lr: LRSplineSurface) -> None:
        super().__init__(lr.w, "nMeshlines", "getMeshline", "meshlineIter", partial(MeshLine, lr))


class MeshRectView(ListLikeView[raw.LRVolume, raw.MeshRectangle, MeshRect]):
    def __init__(self, lr: LRSplineVolume) -> None:
        super().__init__(lr.w, "nMeshRectangles", "getMeshRectangle", "meshrectIter", partial(MeshRect, lr))


class LRSplineObject(ABC, Generic[R]):
    w: R
    elements: ElementView[R]
    basis: BasisView[R]

    @abstractmethod
    def clone(self) -> Self:
        pass

    def __init__(self, w: R, renumber: bool = True) -> None:
        if renumber:
            w.generateIDs()
        self.w = w
        self.elements = ElementView(self)
        self.basis = BasisView(self)

    @staticmethod
    def read_many(stream: BinaryIO | TextIO, renumber: bool = True) -> list[LRSplineObject]:
        contents = stream.read()
        splitter = b"# LRSPLINE" if isinstance(contents, bytes) else "# LRSPLINE"
        contents_list = contents.split(splitter)[1:]  # type: ignore[arg-type]
        contents_list = [splitter + c for c in contents_list]  # type: ignore[operator,assignment]
        return [_constructor(c)(c, renumber=renumber) for c in contents_list]

    @abstractmethod
    def corners(self) -> FArray: ...

    def __len__(self) -> int:
        return len(self.basis)

    @property
    def pardim(self) -> int:
        return next(iter(self.elements)).pardim

    @property
    def dimension(self) -> int:
        return len(next(iter(self.basis)).controlpoint)

    @dimension.setter
    def dimension(self, value: int) -> None:
        self.w.rebuildDimension(value)

    @property
    def shape(self) -> tuple[int]:
        return (len(self),)

    @property
    def controlpoints(self) -> FArray:
        return np.array([bf.controlpoint for bf in self.basis])

    @controlpoints.setter
    def controlpoints(self, value: FArray) -> None:
        if value.shape[0] != len(self):
            raise ValueError(f"Incorrect number of control points: expected {len(self)}")
        if value.ndim != 2:
            raise ValueError("Incorrect number of dimensions: expected 2")
        _, newdim = value.shape
        if newdim != self.dimension:
            self.dimension = newdim
        self.w.setControlpoints(value.flatten())

    def write(self, stream: TextIO) -> None:
        return self.w.write(stream)

    @overload
    def start(self) -> tuple[float, ...]: ...

    @overload
    def start(self, direction: Direction) -> float: ...

    def start(self, direction=None):  # type: ignore[no-untyped-def]
        if direction is None:
            return tuple(self.w.startparam(d) for d in range(self.pardim))
        return self.w.startparam(_check_direction(direction, self.pardim))

    @overload
    def end(self) -> tuple[float, ...]: ...

    @overload
    def end(self, direction: Direction) -> float: ...

    def end(self, direction=None):  # type: ignore[no-untyped-def]
        if direction is None:
            return tuple(self.w.endparam(d) for d in range(self.pardim))
        return self.w.endparam(_check_direction(direction, self.pardim))

    @overload
    def span(self) -> tuple[tuple[float, float], ...]: ...

    @overload
    def span(self, direction: Direction) -> tuple[float, float]: ...

    def span(self, direction=None):  # type: ignore[no-untyped-def]
        if direction is None:
            return tuple((self.w.startparam(d), self.w.endparam(d)) for d in range(self.pardim))
        direction = _check_direction(direction, self.pardim)
        return (self.w.startparam(direction), self.w.endparam(direction))

    @overload
    def order(self) -> tuple[int, ...]: ...

    @overload
    def order(self, direction: Direction) -> int: ...

    def order(self, direction=None):  # type: ignore[no-untyped-def]
        if direction is None:
            return tuple(self.w.order(d) for d in range(self.pardim))
        return self.w.order(_check_direction(direction, self.pardim))

    @overload
    def knots(self, *, with_multiplicities: bool = ...) -> tuple[FArray, ...]: ...

    @overload
    def knots(self, direction: Direction, *, with_multiplicities: bool = ...) -> FArray: ...

    def knots(self, direction=None, with_multiplicities=False):  # type: ignore[no-untyped-def]
        knots = self.w.getGlobalKnotVector() if with_multiplicities else self.w.getGlobalUniqueKnotVector()

        if direction is None:
            return knots
        direction = _check_direction(direction, self.pardim)
        return knots[direction]

    def refine(
        self,
        objects: Sequence[float] | Sequence[BasisFunction] | Sequence[Element],
        beta: float | None = None,
    ) -> None:
        if not objects:
            raise ValueError("Refinement list must be non-empty")
        if isinstance(objects[0], float):
            assert isinstance(beta, float)
            self.w.refineByDimensionIncrease(np.asarray(objects), beta)
        elif isinstance(objects[0], BasisFunction):
            objects = cast("Sequence[BasisFunction]", objects)
            ids = [bf.id for bf in objects]
            self.w.refineBasisFunction(ids)
        elif isinstance(objects[0], Element):
            objects = cast("Sequence[Element]", objects)
            ids = [bf.id for bf in objects]
            self.w.refineElement(ids)
        else:
            raise TypeError("List of unknown objects: expected float, BasisFunction or Element")
        self.w.generateIDs()

    def configure(self, **kwargs: Unpack[ConfigureKwargs]) -> None:
        if "aspect_ratio" in kwargs:
            r = kwargs.pop("aspect_ratio")
            posteriori = kwargs.pop("posteriori_fix", True)
            self.w.setMaxAspectRatio(r, posteriori)
        for key, val in kwargs.items():
            {
                "strategy": "setRefStrat",
                "symmetry": "setRefSymmetry",
                "multiplicity": "setRefMultiplicity",
                "max_tjoints": "setMaxTjoints",
                "close_gaps": "setCloseGaps",
            }[key]
            getattr(self.w, key)(val)

    def generate_ids(self) -> None:
        self.w.generateIDs()

    def bezier_extraction(self, arg: Element | int) -> FArray:
        if isinstance(arg, Element):
            return self.w.getBezierExtraction(arg.id)
        return self.w.getBezierExtraction(arg)  # int

    def __mul__(self, x: Any) -> Self:
        new = self.clone()
        new.controlpoints *= x
        return new

    def __rmul__(self, x: Any) -> Self:
        return self.__mul__(x)

    def __getitem__(self, i: int) -> FArray:
        return self.basis[i].controlpoint

    @abstractmethod
    def element_at(self, *args: float) -> Element: ...

    @abstractmethod
    def __call__(self, *args: FArray | float, iel: int = ...) -> FArray: ...


class LRSplineSurface(LRSplineObject[raw.LRSurface]):
    @overload
    def __init__(self) -> None: ...

    @overload
    def __init__(self, surface: raw.LRSurface) -> None: ...

    @overload
    def __init__(self, data: BinaryIO | TextIO | str | bytes) -> None: ...

    @overload
    def __init__(self, n1: int, n2: int) -> None: ...

    @overload
    def __init__(self, n1: int, n2: int, order_u: int, order_v: int) -> None: ...

    @overload
    def __init__(
        self, n1: int, n2: int, order_u: int, order_v: int, knot1: Scalars, knot2: Scalars
    ) -> None: ...

    @overload
    def __init__(
        self, n1: int, n2: int, order_u: int, order_v: int, knot1: Scalars, knot2: Scalars, cps: ArrayLike
    ) -> None: ...

    def __init__(self, *args, renumber=True):  # type: ignore[no-untyped-def]
        if len(args) == 0:
            w = raw.LRSurface()
        elif isinstance(args[0], raw.LRSurface):
            w = args[0]
        elif isinstance(args[0], io.IOBase | str | bytes):
            w = raw.LRSurface()
            w.read(args[0])  # type: ignore[arg-type]
        elif len(args) == 2:  # specify (n1,n2)
            w = raw.LRSurface(args[0], args[1], 2, 2)
        elif len(args) == 4:  # specify (n1,n2) and (p1,p2)
            w = raw.LRSurface(args[0], args[1], args[2], args[3])
        elif len(args) == 6:  # specify (n1,n2), (p1,p2) and (knot1,knot2)
            w = raw.LRSurface(args[0], args[1], args[2], args[3], args[4], args[5])
        elif len(args) == 7:  # specify controlpoints in addition
            cp = np.asarray(args[-1], dtype=float)
            w = raw.LRSurface(args[0], args[1], args[2], args[3], args[4], args[5], cp.flat, len(cp[0]))
        else:
            w = raw.LRSurface()
        super().__init__(w, renumber=renumber)
        self.meshlines = MeshLineView(self)

    def corners(self) -> FArray:
        return np.array(
            [
                next(self.basis.edge("east", "south")).controlpoint,
                next(self.basis.edge("west", "south")).controlpoint,
                next(self.basis.edge("east", "north")).controlpoint,
                next(self.basis.edge("west", "north")).controlpoint,
            ]
        )

    def clone(self) -> LRSplineSurface:
        return LRSplineSurface(self.w.copy())

    def write_postscript(self, stream: TextIO, **kwargs: Unpack[WritePsKwargs]) -> None:
        return self.w.writePostscriptElements(stream, **kwargs)

    def insert_knot(self, new_knots: ScalarOrScalars, direction: Direction, multiplicity: int = 1) -> None:
        new_knots = _ensure_scalars(new_knots)
        direction = _check_direction(direction, self.pardim)
        for k in new_knots:
            if direction == 0:
                self.w.insert_const_u_edge(k, self.start(1), self.end(1), multiplicity)
            else:
                self.w.insert_const_v_edge(k, self.start(0), self.end(0), multiplicity)
        self.w.generateIDs()

    @overload
    def insert(
        self,
        arg: MeshLine,
        /,
        direction: Direction | None = ...,
        value: float | None = ...,
        start: float | None = ...,
        end: float | None = ...,
        multiplicity: int | None = ...,
    ) -> None: ...

    @overload
    def insert(
        self,
        direction: Direction,
        value: float,
        start: float,
        end: float,
        multiplicity: int | None = ...,
    ) -> None: ...

    def insert(self, *args, direction=None, value=None, start=None, end=None, multiplicity=None):  # type: ignore[no-untyped-def]
        if len(args) > 1:
            raise TypeError("Expected at most one positional argument")

        if len(args) == 1:
            (ml,) = args
            if not isinstance(ml, MeshLine):
                raise TypeError(f"Expected MeshLine, got {type(ml)}")
            if direction is None:
                direction = ml.constant_direction
            if value is None:
                value = ml.value
            if start is None:
                start = ml.start(ml.variable_direction)
            if end is None:
                end = ml.end(ml.variable_direction)
            if multiplicity is None:
                multiplicity = ml.multiplicity

        assert direction is not None
        assert value is not None
        assert start is not None
        assert end is not None

        if multiplicity is None:
            multiplicity = 1
        direction = _check_direction(direction, self.pardim)

        if direction == 0:
            self.w.insert_const_u_edge(value, start, end, multiplicity)
        else:
            self.w.insert_const_v_edge(value, start, end, multiplicity)

    @overload
    def evaluate(self, u: FArray, v: FArray, iel: int = ...) -> FArray: ...

    @overload
    def evaluate(self, u: float, v: float, iel: int = ...) -> FArray: ...

    def evaluate(self, u, v, iel=-1):  # type: ignore[no-untyped-def]
        if isinstance(u, np.ndarray) and isinstance(v, np.ndarray):
            retval = np.array([self.w.point(up, vp, iEl=iel) for up, vp in zip(u.flat, v.flat)])
            return retval.reshape(u.shape + (-1,))
        return self.w.point(u, v, iEl=iel)

    @overload
    def derivative(self, u: FArray, v: FArray, d: tuple[int, int] = ..., iel: int = ...) -> FArray: ...

    @overload
    def derivative(self, u: float, v: float, d: tuple[int, int] = ..., iel: int = ...) -> FArray: ...

    def derivative(self, u, v, d=(1, 1), iel=-1):  # type: ignore[no-untyped-def]
        def wrapper(u: np.floating, v: np.floating, nderivs: int) -> FArray:
            return self.w.point(u, v, nderivs, iEl=iel)

        return _derivative_helper((u, v), d, wrapper)

    def element_at(self, u: float, v: float) -> Element:  # type: ignore[override]
        return self.elements[self.w.getElementContaining(u, v)]

    def make_integer_knots(self) -> float:
        return self.w.makeIntegerKnots()

    __call__ = evaluate  # type: ignore[assignment]


class LRSplineVolume(LRSplineObject[raw.LRVolume]):
    @overload
    def __init__(self) -> None: ...

    @overload
    def __init__(self, surface: raw.LRVolume) -> None: ...

    @overload
    def __init__(self, data: BinaryIO | TextIO | str | bytes) -> None: ...

    @overload
    def __init__(self, n1: int, n2: int, n3: int) -> None: ...

    @overload
    def __init__(self, n1: int, n2: int, n3: int, order_u: int, order_v: int, order_w: int) -> None: ...

    @overload
    def __init__(
        self,
        n1: int,
        n2: int,
        n3: int,
        order_u: int,
        order_v: int,
        order_w: int,
        knot1: Scalars,
        knot2: Scalars,
        knot3: Scalars,
    ) -> None: ...

    @overload
    def __init__(
        self,
        n1: int,
        n2: int,
        n3: int,
        order_u: int,
        order_v: int,
        order_w: int,
        knot1: Scalars,
        knot2: Scalars,
        knot3: Scalars,
        cps: ArrayLike,
    ) -> None: ...

    def __init__(self, *args, renumber=True):  # type: ignore[no-untyped-def]
        if len(args) == 0:
            w = raw.LRVolume()
        elif isinstance(args[0], raw.LRVolume):
            w = args[0]
        elif isinstance(args[0], io.IOBase | str | bytes):
            w = raw.LRVolume()
            w.read(args[0])  # type: ignore[arg-type]
        elif len(args) == 3:  # only specify number of functions (n1,n2,n3)
            w = raw.LRVolume(args[0], args[1], args[2], 2, 2, 2)
        elif len(args) == 6 or len(args) == 9:  # specity n & p for 3 directions
            w = raw.LRVolume(*args)
        elif len(args) == 10:  # specify all above in addition to controlpoints
            cp = np.array(args[-1])
            w = raw.LRVolume(*args[:-1], cp.flatten(), len(cp[0]))  # type: ignore[call-overload]
        else:
            w = raw.LRVolume()
        super().__init__(w, renumber=renumber)
        self.meshrects = MeshRectView(self)

    def corners(self) -> FArray:
        return np.array(
            [
                next(self.basis.edge("east", "south", "bottom")).controlpoint,
                next(self.basis.edge("west", "south", "bottom")).controlpoint,
                next(self.basis.edge("east", "north", "bottom")).controlpoint,
                next(self.basis.edge("west", "north", "bottom")).controlpoint,
                next(self.basis.edge("east", "south", "top")).controlpoint,
                next(self.basis.edge("west", "south", "top")).controlpoint,
                next(self.basis.edge("east", "north", "top")).controlpoint,
                next(self.basis.edge("west", "north", "top")).controlpoint,
            ]
        )

    def clone(self) -> LRSplineVolume:
        return LRSplineVolume(self.w.copy())

    def insert_knot(self, new_knots: ScalarOrScalars, direction: Direction, multiplicity: int = 1) -> None:
        new_knots = _ensure_scalars(new_knots)
        direction = _check_direction(direction, self.pardim)
        for k in new_knots:
            start = list(self.start())
            end = list(self.end())
            start[direction] = end[direction] = k
            mr = raw.MeshRectangle(start[0], start[1], start[2], end[0], end[1], end[2], multiplicity)
            self.w.insert_line(mr)
        self.w.generateIDs()

    def insert(self, mr: MeshRect) -> None:
        self.w.insert_line(mr.w)

    @overload
    def evaluate(self, u: FArray, v: FArray, w: FArray, iel: int = ...) -> FArray: ...

    @overload
    def evaluate(self, u: float, v: float, w: float, iel: int = ...) -> FArray: ...

    def evaluate(self, u, v, w, iel=-1):  # type: ignore[no-untyped-def]
        if isinstance(u, np.ndarray) and isinstance(v, np.ndarray) and isinstance(w, np.ndarray):
            retval = np.array(
                [self.w.point(up, vp, wp, iEl=iel) for up, vp, wp in zip(u.flat, v.flat, w.flat)]
            )
            return retval.reshape(u.shape + (-1,))
        return self.w.point(u, v, w, iEl=iel)

    @overload
    def derivative(
        self, u: FArray, v: FArray, w: FArray, d: tuple[int, int, int] = ..., iel: int = ...
    ) -> FArray: ...

    @overload
    def derivative(
        self, u: float, v: float, w: float, d: tuple[int, int, int] = ..., iel: int = ...
    ) -> FArray: ...

    def derivative(self, u, v, w, d=(1, 1, 1), iel=-1):  # type: ignore[no-untyped-def]
        def wrapper(u: np.floating, v: np.floating, w: np.floating, nderivs: int) -> FArray:
            return self.w.point(u, v, w, nderivs, iEl=iel)

        return _derivative_helper((u, v, w), d, wrapper)

    def element_at(self, u: float, v: float, w: float) -> Element:  # type: ignore[override]
        return self.elements[self.w.getElementContaining(u, v, w)]

    __call__ = evaluate  # type: ignore[assignment]



from pathlib import Path

from Cython.Build import cythonize
from setuptools import Extension


def pdm_build_update_setup_kwargs(context, setup_kwargs):
    lrsplines_path = Path('submodules') / 'LRSplines'
    lrsplines_src_path = lrsplines_path / 'src'
    lrsplines_cpp = [
        lrsplines_src_path / 'Basisfunction.cpp',
        lrsplines_src_path / 'Element.cpp',
        lrsplines_src_path / 'Meshline.cpp',
        lrsplines_src_path / 'MeshRectangle.cpp',
        lrsplines_src_path / 'LRSpline.cpp',
        lrsplines_src_path / 'LRSplineSurface.cpp',
        lrsplines_src_path / 'LRSplineVolume.cpp',
    ]

    extensions = [
        Extension(
            'lrspline.raw',
            ['lrspline/raw.pyx', *map(str, lrsplines_cpp)],
            include_dirs=[str(lrsplines_path / 'include')],
            extra_compile_args=['-std=c++11'],
            define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')],
        ),
    ]

    setup_kwargs.update(
        ext_modules=cythonize(extensions),
        packages=["lrspline"],
    )



from collections.abc import Iterator, Sequence
from typing import BinaryIO, Self, TextIO, overload

import numpy as np
from numpy.typing import ArrayLike, NDArray

class EOFError(Exception): ...

class parameterEdge:
    NONE: int
    WEST: int
    EAST: int
    SOUTH: int
    NORTH: int
    TOP: int
    BOTTOM: int

class refinementStrategy:
    LR_MINSPAN: int
    LR_FULLSPAN: int
    LR_STRUCTURED_MESH: int

class Basisfunction:
    def getId(self) -> int: ...
    def getControlPoint(self) -> NDArray[np.double]: ...
    def nVariate(self) -> int: ...
    @overload
    def evaluate(self, u: np.floating, v: np.floating, u_from_right: bool, v_from_right: bool) -> float: ...
    @overload
    def evaluate(
        self, u: np.floating, v: np.floating, d: int, u_from_right: bool, v_from_right: bool
    ) -> NDArray[np.double]: ...
    @overload
    def evaluate(
        self,
        u: np.floating,
        v: np.floating,
        w: np.floating,
        u_from_right: bool,
        v_from_right: bool,
        w_from_right: bool,
    ) -> float: ...
    @overload
    def evaluate(
        self,
        u: np.floating,
        v: np.floating,
        w: np.floating,
        d: int,
        u_from_right: bool,
        v_from_right: bool,
        w_from_right: bool,
    ) -> NDArray[np.double]: ...
    def supportIter(self) -> Iterator[Element]: ...
    def getknots(self, idx: int) -> NDArray[np.double]: ...

class Element:
    def getId(self) -> int: ...
    def getDim(self) -> int: ...
    def getParmin(self, i: int) -> float: ...
    def getParmax(self, i: int) -> float: ...
    def nBasisFunctions(self) -> int: ...
    def supportIter(self) -> Iterator[Basisfunction]: ...

class Meshline:
    def is_spanning_u(self) -> int: ...
    @property
    def const_par_(self) -> float: ...
    @property
    def start_(self) -> float: ...
    @property
    def stop_(self) -> float: ...
    @property
    def multiplicity_(self) -> int: ...

class MeshRectangle:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(
        self, u0: float, v0: float, w0: float, u1: float, v1: float, w1: float, multiplicity: int = ...
    ) -> None: ...
    def constDirection(self) -> int: ...
    def constParameter(self) -> float: ...
    @property
    def start_(self) -> NDArray[np.double]: ...
    @property
    def stop_(self) -> NDArray[np.double]: ...
    @property
    def multiplicity_(self) -> int: ...
    def copy(self) -> MeshRectangle: ...

class LRSplineObject:
    def nVariate(self) -> int: ...
    def dimension(self) -> int: ...
    def rebuildDimension(self, dim: int) -> None: ...
    def nBasisFunctions(self) -> int: ...
    def nElements(self) -> int: ...
    def startparam(self, i: int) -> float: ...
    def endparam(self, i: int) -> float: ...
    def order(self, i: int) -> int: ...
    def elementIter(self) -> Iterator[Element]: ...
    def getElement(self, i: int) -> Element: ...
    def getEdgeElementsIter(self, edge: int) -> Iterator[Element]: ...
    def basisIter(self) -> Iterator[Basisfunction]: ...
    def getBasifunction(self) -> Iterator[Basisfunction]: ...
    def getEdgeFunctionsIter(self, edge: int, depth: int = ...) -> Iterator[Basisfunction]: ...
    def setControlpoints(self, cpts: NDArray[np.double]) -> None: ...
    def generateIDs(self) -> int: ...
    def refineBasisFunction(self, i: int | list[int]) -> None: ...
    def refineElement(self, i: int | list[int]) -> None: ...
    def refineByDimensionIncrease(self, errors: NDArray[np.double], beta: float) -> None: ...
    def setRefStrat(self, strat: int) -> None: ...
    def setRefSymmetry(self, symmetry: int) -> None: ...
    def setRefMultiplicity(self, mult: int) -> None: ...
    def setMaxTjoints(self, n: int) -> None: ...
    def setCloseGaps(self, close: bool) -> None: ...
    def setMaxAspectRatio(self, r: float, aposteriori: bool = ...) -> None: ...
    def write(self, stream: TextIO) -> None: ...
    def getGlobalKnotVector(self) -> tuple[NDArray[np.double], ...]: ...
    def getGlobalUniqueKnotVector(self) -> tuple[NDArray[np.double], ...]: ...
    def getBezierExtraction(self, iEl: int = ...) -> NDArray[np.double]: ...
    def read(self, stream: BinaryIO | TextIO | bytes | str) -> None: ...
    def copy(self) -> Self: ...

class LRSurface(LRSplineObject):
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, n1: int, n2: int, order_u: int, order_v: int) -> None: ...
    @overload
    def __init__(
        self, n1: int, n2: int, order_u: int, order_v: int, knot1: ArrayLike, knot2: ArrayLike
    ) -> None: ...
    @overload
    def __init__(
        self,
        n1: int,
        n2: int,
        order_u: int,
        order_v: int,
        knot1: ArrayLike,
        knot2: ArrayLike,
        cps: ArrayLike,
        dim: int = ...,
    ) -> None: ...
    def writePostscriptElements(
        self,
        stream: TextIO,
        nu: int = ...,
        nv: int = ...,
        close: bool = ...,
        colorElements: Sequence[int] | None = ...,
    ) -> None: ...
    def insert_const_u_edge(
        self, u: float, start_v: float, stop_v: float, multiplicity: int = ...
    ) -> None: ...
    def insert_const_v_edge(
        self, v: float, start_u: float, stop_u: float, multiplicity: int = ...
    ) -> None: ...
    @overload
    def point(self, u: np.floating, v: np.floating, *, iEl: int = ...) -> NDArray[np.double]: ...
    @overload
    def point(
        self, u: np.floating, v: np.floating, ufr: bool, vfr: bool, *, iEl: int = ...
    ) -> NDArray[np.double]: ...
    @overload
    def point(self, u: np.floating, v: np.floating, d: int, *, iEl: int = ...) -> NDArray[np.double]: ...
    @overload
    def point(
        self, u: np.floating, v: np.floating, d: int, ufr: bool, vfr: bool, *, iEl: int = ...
    ) -> NDArray[np.double]: ...
    def getElementContaining(self, u: float, v: float) -> int: ...
    def makeIntegerKnots(self) -> float: ...
    def meshlineIter(self) -> Iterator[Meshline]: ...
    def nMeshlines(self) -> int: ...
    def getMeshline(self, i: int) -> Meshline: ...

class LRVolume(LRSplineObject):
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, n1: int, n2: int, n3: int, order_u: int, order_v: int, order_w: int) -> None: ...
    @overload
    def __init__(
        self,
        n1: int,
        n2: int,
        n3: int,
        order_u: int,
        order_v: int,
        order_w: int,
        knot1: ArrayLike,
        knot2: ArrayLike,
        knot3: ArrayLike,
    ) -> None: ...
    @overload
    def __init__(
        self,
        n1: int,
        n2: int,
        n3: int,
        order_u: int,
        order_v: int,
        order_w: int,
        knot1: ArrayLike,
        knot2: ArrayLike,
        knot3: ArrayLike,
        cps: ArrayLike,
        dim: int = ...,
    ) -> None: ...
    @overload
    def point(
        self, u: np.floating, v: np.floating, w: np.floating, *, iEl: int = ...
    ) -> NDArray[np.double]: ...
    @overload
    def point(
        self,
        u: np.floating,
        v: np.floating,
        w: np.floating,
        ufr: bool,
        vfr: bool,
        wfr: bool,
        *,
        iEl: int = ...,
    ) -> NDArray[np.double]: ...
    @overload
    def point(
        self, u: np.floating, v: np.floating, w: np.floating, d: int, *, iEl: int = ...
    ) -> NDArray[np.double]: ...
    @overload
    def point(
        self,
        u: np.floating,
        v: np.floating,
        w: np.floating,
        d: int,
        ufr: bool,
        vfr: bool,
        wfr: bool,
        *,
        iEl: int = ...,
    ) -> NDArray[np.double]: ...
    def insert_line(self, mr: MeshRectangle) -> None: ...
    def getElementContaining(self, u: float, v: float, w: float) -> int: ...
    def meshrectIter(self) -> Iterator[MeshRectangle]: ...
    def nMeshRectangles(self) -> int: ...
    def getMeshRectangle(self, i: int) -> MeshRectangle: ...
